@@ -21,11 +21,13 @@
 #define ARROW_INTERVAL 100	/* arrow update interval in milliseconds */
 #define SCROLL_INTERVAL 300	/* scroll update interval in milliseconds */
 #define SEPARATOR " ** "	/* The separator for the scrolling title */
+#define DISPLAYSIZE 6		/* width of text to display (running title) */
 
 #include <dockapp.h>
 #include <unistd.h>
 #include <ctype.h>
 #include <string.h>
+#include <locale.h>
 #include <xmms/xmmsctrl.h>
 
 #include "version.h"
@@ -48,6 +50,7 @@ void ActionFastr(int x, int y, DARect rect, void *data);
 void ActionFastf(int x, int y, DARect rect, void *data);
 void ToggleWins(int x, int y, DARect rect, void *data);
 void ToggleVol(int x, int y, DARect rect, void *data);
+void ChangeVol(int x, int y, DARect rect, void *data);
 void ToggleTitle(int x, int y, DARect rect, void *data);
 void ToggleTime(int x, int y, DARect rect, void *data);
 void buttonPress(int button, int state, int x, int y);
@@ -73,12 +76,14 @@ XEvent ev;
 /* Dockapp variables */
 unsigned int xmms_session = 0;
 char *xmms_cmd = "xmms";
-Bool pl_vis=0, eq_vis=0;
+Bool main_vis=0, pl_vis=0, eq_vis=0;
 Bool xmms_running;
 unsigned int volume_step = 5;
+Bool run_excusive=0;
 
 Bool t_scroll=0;
 Bool t_time=0;
+Bool t_volume=0;
 float title_pos = 0;
 unsigned int arrow_pos = 0;
 Bool pause_norotate = 0;
@@ -87,6 +92,9 @@ Time click_time=0;
 Pixmap pixmap, mask;	/* Pixmap that is displayed */
 Pixmap pixnum, masknum;	/* Pixmap source */
 
+int left_pressed = 0; /* for pseudo drag callback */
+int motion_event = 0; /* on motion events we do not want(too fast) display update */
+
 static DAActionRect buttonRects[] = {
 	{{5, 39, 14, 9}, ActionPrev},
 	{{19, 39, 14, 9}, ActionNext},
@@ -107,8 +115,12 @@ static DAActionRect globRect[] = {
 };
 
 static DAActionRect displayRects[] = {
-	{{5, 5, 54, 15}, ToggleTime},
-	{{5, 19, 54, 15}, ToggleTitle}
+	{{5, 5, 54, 12}, ToggleTime},
+	{{5, 25, 54, 9}, ToggleTitle}
+};
+
+static DAActionRect volumeRects[] = {
+	{{5, 17, 38, 8 }, ChangeVol}
 };
 
 static DAProgramOption options[] = {
@@ -129,7 +141,108 @@ static DAProgramOption options[] = {
 	{"-a", "--rotate-arrow", "Do not rotate the arrow, when paused", 
 		DONone, False, {NULL} },
 	{"-l", "--time-left", "Show time left instead of time remaining by default",
-	        DONone, False, {NULL} }
+	 DONone, False, {NULL} },
+	{"-b", "--volume-slider", "Show volume slider instead of channels", 
+	 DONone, False, {NULL}},
+	{"-R", "--run-excusive", "Run xmms on startup, exit when xmms exits", DONone, False, {NULL} }
+};
+
+typedef struct
+{
+	wchar_t c;
+	int x;
+	int y;
+} glyphdescr;
+
+static glyphdescr glyphs[] = {
+	{L'-', 67, 83}, {L'.', 73, 83}, {L'\x27', 79, 83},
+	{L'(', 85, 83}, {L')', 91, 83}, {L'*', 97, 83}, {L'/', 103, 83},
+
+	{L'0',  1, 83}, {L'1',  7, 83}, {L'2', 13, 83}, {L'3', 19, 83}, {L'4', 25, 83},
+	{L'5', 31, 83}, {L'6', 37, 83}, {L'7', 43, 83}, {L'8', 49, 83}, {L'9', 55, 83},
+
+
+
+	{L'A',  1, 73}, {L'a',  1, 73},
+	{L'B',  7, 73}, {L'b',  7, 73},
+	{L'C', 13, 73}, {L'c', 13, 73},
+	{L'D', 19, 73}, {L'd', 19, 73},
+	{L'E', 25, 73}, {L'e', 25, 73},
+
+	{L'F', 31, 73}, {L'f', 31, 73},
+	{L'G', 37, 73}, {L'g', 37, 73},
+	{L'H', 43, 73}, {L'h', 43, 73},
+	{L'I', 49, 73}, {L'i', 49, 73},
+	{L'J', 55, 73}, {L'j', 55, 73},
+
+	{L'K', 61, 73}, {L'k', 61, 73},
+	{L'L', 67, 73}, {L'l', 67, 73},
+	{L'M', 73, 73}, {L'm', 73, 73},
+	{L'N', 79, 73}, {L'n', 79, 73},
+	{L'O', 85, 73}, {L'o', 85, 73},
+
+	{L'P', 91, 73}, {L'p', 91, 73},
+	{L'Q', 97, 73}, {L'q', 97, 73},
+	{L'R',103, 73}, {L'r',103, 73},
+	{L'S',109, 73}, {L's',109, 73},
+	{L'T',115, 73}, {L't',115, 73},
+
+	{L'U',121, 73}, {L'u',121, 73},
+	{L'V',127, 73}, {L'v',127, 73},
+	{L'W',133, 73}, {L'w',133, 73},
+	{L'X',139, 73}, {L'x',139, 73},
+	{L'Y',145, 73}, {L'y',145, 73},
+
+	{L'Z',151, 73}, {L'z',151, 73},
+
+
+	{L'\x42e',  1, 93}, {L'\x44e',  1, 93}, /* cyrillic Yu */
+
+	{L'\x410',  7, 93}, {L'\x430',  7, 93}, /* cyrillic A */
+	{L'\x411', 13, 93}, {L'\x431', 13, 93}, /* cyrillic Be */
+	{L'\x426', 19, 93}, {L'\x446', 19, 93}, /* cyrillic Ce */
+	{L'\x414', 25, 93}, {L'\x434', 25, 93}, /* cyrillic De */
+	{L'\x415', 31, 93}, {L'\x435', 31, 93}, /* cyrillic Ye */
+
+	{L'\x424', 37, 93}, {L'\x444', 37, 93}, /* cyrillic eF */
+	{L'\x413', 43, 93}, {L'\x433', 43, 93}, /* cyrillic Ge */
+	{L'\x425', 49, 93}, {L'\x445', 49, 93}, /* cyrillic Ha */
+	{L'\x418', 55, 93}, {L'\x438', 55, 93}, /* cyrillic I */
+	{L'\x419', 61, 93}, {L'\x439', 61, 93}, /* cyrillic I-kratkoe */
+
+	{L'\x41a', 67, 93}, {L'\x43a', 67, 93}, /* cyrillic Ka */
+	{L'\x41b', 73, 93}, {L'\x43b', 73, 93}, /* cyrillic eL */
+	{L'\x41c', 79, 93}, {L'\x43c', 79, 93}, /* cyrillic eM */
+	{L'\x41d', 85, 93}, {L'\x43d', 85, 93}, /* cyrillic eN */
+	{L'\x41e', 91, 93}, {L'\x43e', 91, 93}, /* cyrillic O */
+
+	{L'\x41f', 97, 93}, {L'\x43f', 97, 93}, /* cyrillic Pe */
+	{L'\x42f',103, 93}, {L'\x44f',103, 93}, /* cyrillic Ya */
+	{L'\x420',109, 93}, {L'\x440',109, 93}, /* cyrillic eR */
+	{L'\x421',115, 93}, {L'\x441',115, 93}, /* cyrillic eS */
+	{L'\x422',121, 93}, {L'\x442',121, 93}, /* cyrillic Te */
+
+	{L'\x423',127, 93}, {L'\x443',127, 93}, /* cyrillic U */
+	{L'\x416',133, 93}, {L'\x436',133, 93}, /* cyrillic Je */
+	{L'\x412',139, 93}, {L'\x432',139, 93}, /* cyrillic Ve */
+	{L'\x42c',145, 93}, {L'\x44c',145, 93}, /* cyrillic MyagkijZnak */
+	{L'\x42b',151, 93}, {L'\x44b',151, 93}, /* cyrillic Y */
+
+	{L'\x417',157, 93}, {L'\x437',157, 93}, /* cyrillic Ze */
+	{L'\x428',163, 93}, {L'\x448',163, 93}, /* cyrillic Sha */
+	{L'\x42d',169, 93}, {L'\x44d',169, 93}, /* cyrillic E */
+	{L'\x429',175, 93}, {L'\x449',175, 93}, /* cyrillic Scha */
+	{L'\x427',181, 93}, {L'\x447',181, 93}, /* cyrillic Che */
+
+	{L'\x42a',187, 93}, {L'\x44a',187, 93}, /* cyrillic TvyordyiZnak */
+	{L'\x404',115, 83}, {L'\x454',115, 83}, /* ukrainian IE */
+	{L'\x406', 49, 73}, {L'\x456', 49, 73}, /* ukrainian I */
+	{L'\x407',109, 83}, {L'\x457',109, 83}, /* ukrainian YI */
+	{L'\x491', 43, 93}, {L'\x490', 43, 93}, /* ukrainian GHE with upturn */
+
+	{L'\x401',121, 83}, {L'\x451',121, 83}, /* cyrillic Yo */
+
+	{L' ', 61, 83}
 };
 
 /*----------------------------------------------------------------------------*/
@@ -235,8 +348,11 @@ void ToggleWins(int x, int y, DARect rect, void *data)
 			xmms_remote_pl_win_toggle(xmms_session,
 					!xmms_remote_is_pl_win(xmms_session));
 		} else {
-			if (xmms_remote_is_main_win(xmms_session))
+			if (xmms_remote_is_main_win(xmms_session) || 
+			    xmms_remote_is_pl_win(xmms_session) ||
+			    xmms_remote_is_eq_win(xmms_session))
 			{
+				main_vis = xmms_remote_is_main_win(xmms_session);
 				pl_vis = xmms_remote_is_pl_win(xmms_session);
 				eq_vis = xmms_remote_is_eq_win(xmms_session);
 
@@ -244,7 +360,7 @@ void ToggleWins(int x, int y, DARect rect, void *data)
 				xmms_remote_pl_win_toggle(xmms_session, 0);
 				xmms_remote_eq_win_toggle(xmms_session, 0);
 			} else {
-				xmms_remote_main_win_toggle(xmms_session, 1);
+				xmms_remote_main_win_toggle(xmms_session, main_vis);
 				xmms_remote_pl_win_toggle(xmms_session, pl_vis);
 				xmms_remote_eq_win_toggle(xmms_session, eq_vis);
 			}
@@ -276,6 +392,12 @@ void ToggleVol(int x, int y, DARect rect, void *data)
 	xmms_remote_set_main_volume(xmms_session, volume+factor);
 }
 
+void ChangeVol(int x, int y, DARect rect, void *data)
+{
+	int volume = (int)100*(((float)x)/38);
+	xmms_remote_set_main_volume(xmms_session, volume);
+}
+
 void ToggleTitle(int x, int y, DARect rect, void *data)
 {
 	if (t_scroll)
@@ -292,6 +414,8 @@ void ToggleTime(int x, int y, DARect rect, void *data)
 
 void buttonPress(int button, int state, int x, int y)
 {
+	if (button==1)
+		left_pressed=1;
 	if (xmms_running)
 	{
 		if (button == 1)
@@ -329,6 +453,8 @@ void buttonPress(int button, int state, int x, int y)
 
 void buttonRelease(int button, int state, int x, int y)
 {
+	if (button==1)
+		left_pressed=0;
 	if (xmms_running)
 	{
 		if (button == 1)
@@ -340,17 +466,31 @@ void buttonRelease(int button, int state, int x, int y)
 	}
 }
 
+void buttonDrag(int x, int y) 
+{
+	motion_event=1;
+	if (t_volume)
+		if (left_pressed==1) {
+			DAProcessActionRects(x,y, volumeRects, 1, NULL);
+			DrawChannels(0); /* volume only update */
+			DASetPixmap(pixmap);
+		}
+}
+
 void DisplayRoutine()
 {
-	int time, length, position;
+	int time = 0, length = 0, position = 100;
 	int rate, freq, nch;
 	char *title = NULL;
 	Bool do_scroll=t_scroll;
+
 	xmms_running=xmms_remote_is_running(xmms_session);
 
 	/* Compute diplay */
 	if (!xmms_running)
 	{
+		if (run_excusive)
+			exit(0);
 		position = 100; /* Will display "00" as the current position */
 		do_scroll = 0;   /* Will display channel/kbps info */
 		arrow_pos = 5;
@@ -360,14 +500,19 @@ void DisplayRoutine()
 		freq = 0;
 		length = 0;
 	} else {
-		position = xmms_remote_get_playlist_pos(xmms_session);
-		xmms_remote_get_info(xmms_session, &rate, &freq, &nch);
-		time = xmms_remote_get_output_time(xmms_session);
-		length = xmms_remote_get_playlist_time(xmms_session, position);
-		title = xmms_remote_get_playlist_title(xmms_session, position);
-		if (!xmms_remote_is_playing(xmms_session) || (pause_norotate && 
-					xmms_remote_is_paused(xmms_session) ))
-			arrow_pos = 5;
+	        /* check again, avoiding xmms_remote_* functions to hang,
+		   if xmms was closed in the meantime */
+	        if ((xmms_running = xmms_remote_is_running(xmms_session)))  
+		{
+		        position = xmms_remote_get_playlist_pos(xmms_session);
+			xmms_remote_get_info(xmms_session, &rate, &freq, &nch);
+			time = xmms_remote_get_output_time(xmms_session);
+			length = xmms_remote_get_playlist_time(xmms_session, position);
+			title = xmms_remote_get_playlist_title(xmms_session, position);
+			if (!xmms_remote_is_playing(xmms_session) || (pause_norotate && 
+								      xmms_remote_is_paused(xmms_session) ))
+			    arrow_pos = 5;
+		}
 	}
 
 	/*Draw everything */
@@ -446,11 +591,20 @@ void DrawArrow(void)
 
 void DrawChannels(int channels)
 {
-	if (channels == 4)
-	{
-		copyNumArea(55, 32, 10, 6, 10, 18);
+
+	if (t_volume) {
+   		int volume = (int)(0.36*(float)xmms_remote_get_main_volume(xmms_session));
+		if (volume > 36)
+			volume = 36;
+		copyNumArea(61, 0, volume, 6, 7, 18);
+		copyNumArea(97, 0, 36-volume, 6, 7+volume, 18);
 	} else {
-		copyNumArea((channels*10), 22, 10, 6, 10, 18);
+		if (channels == 4)
+			{
+				copyNumArea(55, 32, 10, 6, 10, 18);
+			} else {
+				copyNumArea((channels*10), 22, 10, 6, 10, 18);
+			}
 	}
 }
 
@@ -471,103 +625,59 @@ void DrawKbps(int bps)
 	copyNumArea(55, 39, 18, 8, 25, 26);
 }
 
+void DrawChar(wchar_t wc, int x, int y)
+{
+	int i;
+	for(i = 0; i < sizeof(glyphs)/sizeof(glyphdescr) - 1; ++i)
+	{
+		if(wc == glyphs[i].c)
+			break;
+	}
+	copyNumArea(glyphs[i].x, glyphs[i].y, 6, 8, x, y);
+}
+
+int DrawChars(char *title, int tpos, int pos)
+{
+	wchar_t wc;
+
+	mbtowc(NULL, NULL, 0);
+	while(*title && (pos <= (tpos + DISPLAYSIZE)))
+	{
+		int len = mbtowc(&wc, title, MB_CUR_MAX);
+		title += len;
+		if(pos > tpos)
+			DrawChar(wc, (pos - tpos)*6 + 1, 26);
+		++pos;
+	}
+	return pos;
+}
+
 void DrawTitle(char *name)
 {
-	int i, len, max, pos;
-	char *title = NULL;
-	char c=' ';
+	int len, pos, tpos = title_pos;
 
 	if (name == NULL)
 		return;
 #ifdef DEBUG
 	printf("name: %s\n", name);
 #endif
-
-	pos = title_pos;
-
-	len = strlen(name);
-	if (len <6)
+	len = pos = DrawChars(name, tpos, 0);
+	if(pos < 6)
 	{
-		max = len;
-		title = g_strdup(name);
-	} else {
-		max = 6;
-		title = g_strdup_printf("%s%s", name, SEPARATOR);
-		len = strlen(title);
+		DrawChars("      ", tpos, pos);
+		return;
 	}
-	if (pos >= len)
-	{
+
+	if(pos <= tpos + DISPLAYSIZE)
+		pos = DrawChars(SEPARATOR, tpos, pos);
+
+	if(pos <= tpos + DISPLAYSIZE)
+		DrawChars(name, tpos, pos);
+
+	if(tpos >= len + strlen(SEPARATOR))
 		title_pos = 0;
-		pos = 0;
-	}
 
-	/* Start by printing all the characters we can */
-	for (i=0;i<max; i++)
-	{
-		{
-			int tmp = i+pos;
-
-			if (tmp<len)
-				c = toupper(title[tmp]);
-			else
-				c = toupper(title[(tmp-len)%6]);
-		}
-#ifdef DEBUG
-		printf("%c", c);
-#endif
-
-		if (isalpha(c))
-			copyNumArea((c-'A')*6 + 1, 73, 6, 8, (i*6)+7, 26);
-		else if (isdigit(c))
-			copyNumArea((c-'0')*6 + 1, 83, 6, 8, (i*6)+7, 26);
-		else if (isspace(c))
-			copyNumArea(61, 83, 6, 8, (i*6)+7, 26);
-		else 
-		{
-			switch (c)
-			{
-				case '-':
-					copyNumArea(67, 83, 6, 8, (i*6)+7, 26);
-					break;
-				case '.':
-					copyNumArea(73, 83, 6, 8, (i*6)+7, 26);
-					break;
-				case 0x27:
-					copyNumArea(79, 83, 6, 8, (i*6)+7, 26);
-					break;
-				case '(':
-					copyNumArea(85, 83, 6, 8, (i*6)+7, 26);
-					break;
-				case ')':
-					copyNumArea(91, 83, 6, 8, (i*6)+7, 26);
-					break;
-				case '*':
-					copyNumArea(97, 83, 6, 8, (i*6)+7, 26);
-					break;
-				case '/':
-					copyNumArea(103, 83, 6, 8, (i*6)+7, 26);
-					break;
-				default:
-					/* Char not recognized, use a space */
-					copyNumArea(61, 83, 6, 8, (i*6)+7, 26);
-					break;
-			}
-		}
-	}
-#ifdef DEBUG
-	printf("\n");
-#endif
-
-	/* If the title is too short, fill up with spaces */
-	for (;i<6; i++)
-	{
-		copyNumArea(61, 83, 6, 8, (i*6)+7, 26);
-	}
-	/* You don't want a 2 letters title to scroll, do you ? */
-	if (len > 6)
-		title_pos = title_pos + 0.5;
-	/* Clean up */
-	free(title);
+	title_pos = title_pos + 0.5;
 }
 
 void ExecuteXmms(void)
@@ -594,7 +704,7 @@ void ExecuteXmms(void)
 int main(int argc, char **argv)
 {
 	unsigned height, width;
-	DACallbacks callbacks={NULL, buttonPress, buttonRelease, NULL,
+	DACallbacks callbacks={NULL, buttonPress, buttonRelease, buttonDrag,
 		NULL, NULL, NULL};
 
 	/* Initialization */
@@ -603,6 +713,7 @@ int main(int argc, char **argv)
 		"XMMS remote control by Bastien Nocera <hadess@hadess.net>",
 		VERSION);
 
+	setlocale(LC_ALL, "");
 	DAInitialize(displayName, "wmusic", 64, 64, argc, argv);
 	DASetCallbacks(&callbacks);
 	DASetTimeout(100);
@@ -622,10 +733,12 @@ int main(int argc, char **argv)
 	if (options[4].used) t_scroll=1;
 	if (options[8].used) pause_norotate=1;
 	if (options[9].used) t_time=1;
+	if (options[10].used) t_volume=1;
+	if (options[11].used) run_excusive=1;
 
-	/* Launch xmms if it's not running and -r was used */
+	/* Launch xmms if it's not running and -r or -R was used */
 	xmms_running=xmms_remote_is_running(xmms_session);
-	if ((!xmms_running) && (options[2].used))
+	if ((!xmms_running) && (options[2].used || run_excusive))
 	{
 		ExecuteXmms();
 		xmms_running=1;
@@ -640,7 +753,10 @@ int main(int argc, char **argv)
 	while (1) {
 		if (DANextEventOrTimeout(&ev, 100))
 				DAProcessEvent(&ev);
-	DisplayRoutine();
+		if (!motion_event)
+			DisplayRoutine();
+		else
+			motion_event=0;
 	}
 	return 0;
 }
